import copy
import re

from . import basic_parser
from bs4 import BeautifulSoup

SECTION_ID = {
    'noun': 'Noun',
    'verb': 'Verb',
    'adverb': 'Adverb',
    'adjective': 'Adjective',
    'preposition': 'Preposition',
    'numeral': 'Numeral',
    'proper_noun': 'Proper_noun',
    'pronoun': 'Pronoun',
    'particle': 'Particle',
    'interjection': 'Interjection',
    'phrase': 'Phrase',
    'contraction': 'Contraction',
    'conjunction': 'Conjunction',
    'article': 'Article',
    'suffix': 'Suffix',
    'abbreviation': 'Abbreviation',
    'initialism': 'Initialism',
    'proverb': 'Proverb',
    'prefix': 'Prefix'
}


class EnglishParser(basic_parser.BasicParser):
    """Parser for english wiktionary pages"""
    def parse(self):
        """Parse the page to return word attributes. This is an extension of BasicParser

        :return: Word attributes. With first release this contain just the 'meaning' key
        :raises ValueError: Raise a ValueError exception if the parser has an empty invalid html
        """
        if not self.html:
            raise ValueError()
        word = {'meanings': self.get_meanings()}
        return word

    def get_meanings(self, get_examples=True, get_empty_meaning_types=False, language_tag_id='English'):
        return self._get_meanings(self._get_language_soup(language_tag_id), get_examples, get_empty_meaning_types)

    def _get_meanings(self, soup, get_examples=True, get_empty_meaning_types=False):
        meanings = {}
        if soup:
            soup_copy = copy.copy(soup)
            meaning_list_tags = soup_copy.find_all('ol')
            for ol_tag in meaning_list_tags:
                meaning_list = []
                type_tag = ol_tag.previous_element
                type_readable_name = type_tag.get_text().strip()
                while type_tag and (not type_tag.name or (not type_tag.name.startswith('h') and not type_tag.name == 'p')):
                    type_tag = type_tag.previous_element
                if type_tag.name == 'p':
                    while type_tag and (not type_tag.name or not type_tag.name.startswith('h')):
                        type_tag = type_tag.previous_element
                    type_name = type_tag.get_text().lower()
                    for li_meaning in ol_tag.find_all('li', recursive=False):
                        examples = []
                        meaning_list.append({
                            'meaning':  re.sub(r'\s+', ' ', li_meaning.text).strip(),
                            'examples': examples
                        })
                    meanings[type_name] = {'name': type_readable_name, 'meanings': meaning_list}
        return meanings

    def get_meaning_by_list(self, meaning_types=list(SECTION_ID.keys()), get_examples=True, get_empty_meaning_types=False):
        """Get english meanings of the word find in the Wiktionary HTML page contained in this parser

        :param list meaning_types: List containing meaning type keys to looking for. By default this list contain all foundable keys
        :param boolean get_examples: Specify if the result should contain example of meanings. By default this is True
        :param boolean get_empty_meaning_types: Specify if empty meaning type keys should be included in the returned wiktionary. By default this is False
        :return dictionary: Return dictionary with all found meanings divided by meaning type keys. Each entry contain a list of dictionaries with two entry: 'meaning' which is a meaning of the word, and 'example' with a list of examples, if are present and should be returned, for the retrieved meaning.
        """
        return self._get_meanings_by_list(self._get_language_soup('English'), meaning_types, get_examples,
                                  get_empty_meaning_types)

    def _get_meaning_by_list(self, soup, meaning_types=list(SECTION_ID.keys()), get_examples=True, get_empty_meaning_types=False):
        """Private function for public get_meanings method that get the soup parser as input with the other passed parameters of get_meanings function"""
        meanings = {}
        if get_empty_meaning_types:
            for meaning_type in list(SECTION_ID.keys()):
                meanings[meaning_type] = []
        if soup:
            soup_copy = copy.copy(soup)
            for meaning_type in meaning_types:
                if meaning_type in SECTION_ID:
                    meaning_list = []
                    title = soup_copy.find(id=SECTION_ID[meaning_type])
                    if title:
                        for li_meaning in title.find_next('ol').find_all('li', recursive=False):
                            examples = []
                            example_list = li_meaning.find('dl')
                            if example_list:
                                example_list = example_list.extract()
                                if get_examples:
                                    for li_example in example_list.find_all('dd'):
                                        if li_example.text.rstrip():
                                            examples.append(re.sub(r'\s+', ' ', li_example.text).strip())
                            meaning_list.append({
                                'meaning':  re.sub(r'\s+', ' ', li_meaning.text).strip(),
                                'examples': examples
                            })
                    if get_empty_meaning_types or meaning_list:
                        meanings[meaning_type] = meaning_list
        return meanings

    def _get_language_soup(self, language_section_id):
        """Private function to get sub-soup parser just for the specified language looked for its code or id in the HTML wiktionary page"""
        language_tag = self.soup.find(id=language_section_id)
        if language_tag:
            english_section_html = ''
            for el in self.soup.find(id=language_section_id).parent.next_siblings:
                if el.name == 'h2':
                    break
                english_section_html += str(el)
            return BeautifulSoup(english_section_html, 'html.parser')
        else:
            return None



import requests
import pywiktionary
import urllib.parse
from pywiktionary import parsers


"""Language code by language"""
LANGUAGE_CODES = {
    'italian': 'it',
    'it': 'it',
    'english': 'en',
    'en': 'en'
}

"""Language Parser to use by default be language code"""
LANGUAGE_PARSERS = {
    'it': parsers.ItalianParser,
    'en': parsers.EnglishParser
}


class WiktionaryParserFactory:
    """Class to retrieve wiktionary pages to parse based on passed language"""
    def __init__(self, default_language='en', default_parser_class=None):
        """Init the parser factory specifying language to use and, if necessary, the parser class to use

        :param string default_language: Language to use to decide which wiktionary language use. Default is 'en' to use English language
        :param default_parser_class: If it's necessary this specify which class to use to instantiate a wiktionary page parser. By default this is None to use the parser class associated to the passed language. If the passed language is not support yet, you should implement a parser and pass the class here
        """
        # TODO: Raise an exception when a language not implemented yet has been passed AND the default parser class is not passed
        self._base_url = 'https://{lang_code}.wiktionary.org/w/index.php?printable=yes&title={page_title}'
        self.default_language = LANGUAGE_CODES[default_language] if default_language in LANGUAGE_CODES else default_language
        if default_parser_class:
            self.default_parser_class = default_parser_class
        else:
            if self.default_language in LANGUAGE_PARSERS:
                self.default_parser_class = LANGUAGE_PARSERS[self.default_language]
            else:
                self.default_parser_class = parsers.BasicParser

    def get_page(self, title, language=None):
        """Get a wiktionary page and

        :param string title: Title or word of the wiktionary page to retrieve and parse
        :param language: Override default language with desired language code to retrieve a wiktionary page from another language
        :return: Parser class of selected language
        """
        if not language or language not in LANGUAGE_CODES:
            language = self.default_language
        wiktionary_response = requests.get(self._format_url(language=language, title=title))
        if wiktionary_response.status_code == 404:
            raise PageNotFoundException
        return self.default_parser_class(wiktionary_response.text)

    def _format_url(self, language, title):
        """Format the base wiktionary URL with language and requested page

        :param language: Language to use in wiktionary site
        :param title: Title of page to retrieve
        :return string: Formatted wiktionary URL
        """
        return self._base_url.format(lang_code=language, page_title=urllib.parse.quote_plus(title))


class PageNotFoundException(requests.RequestException):
    """Exception return when a not found page has been looked for"""
    pass
